@@ -9,8 +9,6 @@ from .utils import POST
 from .utils import FutureResults
 from .utils import enhanceEvent
 
-import gevent
-
 class Sensor( object ):
     '''Representation of a limacharlie.io Sensor.'''
 
@@ -28,7 +26,7 @@ class Sensor( object ):
     _ARCHITECTURE_ALPINE64 = 0x00000005
     _ARCHITECTURE_CHROME = 0x00000006
 
-    def __init__( self, manager, sid ):
+    def __init__( self, manager, sid, detailedInfo = None ):
         try:
             uuid.UUID( sid )
         except:
@@ -40,6 +38,7 @@ class Sensor( object ):
         self._platform = None
         self._architecture = None
         self._hostname = None
+        self._detailedInfo = detailedInfo
 
     def setInvId( self, inv_id ):
         '''Set an investigation ID to be applied to all actions done using the object.
@@ -64,7 +63,7 @@ class Sensor( object ):
         while not self.isOnline():
             if time.time() >= deadline:
                 return False
-            gevent.sleep( min( 60, deadline - time.time() ) )
+            time.sleep( min( 60, deadline - time.time() ) )
 
         return True
 
@@ -133,10 +132,11 @@ class Sensor( object ):
         # Although getting the command result may take a while, the receipt from the sensor
         # should come back quickly so we will implement a static wait for that.
         while True:
-            gevent.sleep( 1 )
+            time.sleep( 1 )
             if future.wasReceived:
                 break
             if time.time() > deadline:
+                print("DEADLINE")
                 return None
 
         # We know the sensor got the tasking, now we will wait according to variable timeout.
@@ -166,18 +166,20 @@ class Sensor( object ):
                     return allResponses
         return None
 
-    def tag( self, tag, ttl ):
+    def tag( self, tag, ttl=None ):
         '''Apply a Tag to the Sensor.
 
         Args:
-            tag (str): Tag to apply.
+            tag (str or list of str): Tag(s) to apply.
             ttl (int): number of seconds the Tag should remain applied.
 
         Returns:
             the REST API response (JSON).
         '''
-
-        req = { 'tags' : str( tag ), 'ttl' : int( ttl ) }
+        if ttl is not None:
+            req = { 'tags' : tag, 'ttl' : int( ttl ) }
+        else:
+            req = { 'tags' : tag }
         return self._manager._apiCall( '%s/tags' % self.sid, POST, req )
 
     def untag( self, tag ):
@@ -190,7 +192,10 @@ class Sensor( object ):
             the REST API response (JSON).
         '''
 
-        req = { 'tag' : str( tag ) }
+        if isinstance( tag, str ):
+            req = { 'tag' : tag }
+        else:
+            req = { 'tags' : ','.join( tag ) }
         return self._manager._apiCall( '%s/tags' % self.sid, DELETE, req )
 
     def getTags( self ):
@@ -209,7 +214,9 @@ class Sensor( object ):
         Returns:
             high level information on the Sensor.
         '''
-        data = self._manager._apiCall( '%s' % self.sid, GET )
+        data = self._detailedInfo
+        if not data:
+            data = self._manager._apiCall( '%s' % self.sid, GET )[ 'info' ]
 
         # We massage the info a bit to make it easier to understand.
         platToString = {
@@ -228,13 +235,14 @@ class Sensor( object ):
             self._ARCHITECTURE_ALPINE64 : 'alpine64',
             self._ARCHITECTURE_CHROME : 'chrome',
         }
-        data = data[ 'info' ]
 
         self._platform = data[ 'plat' ]
         self._architecture = data[ 'arch' ]
         self._hostname = data.get( 'hostname', None )
         self._is_isolated = data.get( 'is_isolated', None )
         self._should_isolate = data.get( 'should_isolate', None )
+        self._is_sealed = data.get( 'is_sealed', None )
+        self._should_seal = data.get( 'should_seal', None )
 
         data[ 'plat' ] = platToString.get( data[ 'plat' ], data[ 'plat' ] )
         data[ 'arch' ] = archToString.get( data[ 'arch' ], data[ 'arch' ] )
@@ -318,7 +326,7 @@ class Sensor( object ):
             self.getInfo()
         return self._hostname
 
-    def getHistoricEvents( self, start, end, limit = None, eventType = None, isForward = True ):
+    def getHistoricEvents( self, start, end, limit = None, eventType = None, isForward = True, outputName = None ):
         '''Get the events for this sensor between the two times, requires Insight (retention) enabled.
 
         Args:
@@ -327,6 +335,7 @@ class Sensor( object ):
             limit (int): maximum number of events to return.
             eventType (str): return events only of this type.
             isForward (bool): return events in ascending order.
+            outputName (str): send data to a named output instead.
 
         Returns:
             a generator of events.
@@ -349,6 +358,10 @@ class Sensor( object ):
 
         if eventType is not None:
             req[ 'event_type' ] = eventType
+        if outputName is not None:
+            req[ 'output_name' ] = outputName
+            yield self._manager._apiCall( 'insight/%s/%s' % ( self._manager._oid, self.sid ), GET, queryParams = req )
+            return
 
         nReturned = 0
         while cursor:
@@ -431,6 +444,36 @@ class Sensor( object ):
 
         return data
 
+    def getChildrenEvents( self, atom ):
+        '''Get all children events from a given atom.
+
+        Args:
+            atom (string): atom to get the children of.
+
+        Returns:
+            List of events.
+        '''
+        req = {
+            'oid' : self._manager._oid,
+            'is_compressed' : True,
+        }
+
+        data = self._manager._apiCall( 'insight/%s/%s/%s/children' % ( self._manager._oid, self.sid, atom ), GET, queryParams = req )
+        data = self._manager._unwrap( data[ 'events' ] )
+
+        return data
+
+    def getEventByAtom( self, atom ):
+        '''Get an event by atom.
+
+        Args:
+            atom (string): atom to get the event of.
+
+        Returns:
+            Event.
+        '''
+        return self._manager._apiCall( 'insight/%s/%s/%s' % ( self._manager._oid, self.sid, atom ), GET )
+
     def delete( self ):
         '''Delete the sensor. It will not be able to connect to the cloud anymore, but will not be uninstalled.abs
         '''
@@ -455,6 +498,40 @@ class Sensor( object ):
         '''Remove the sensor from network isolation (persistent).'''
         return self._manager._apiCall( '%s/isolation' % ( self.sid, ), DELETE )
 
+    def isSealed( self ):
+        '''Determine if the given sensor is marked to be sealed.
+
+        Returns:
+            True if sealed.
+        '''
+        # Seal is ephemeral, so always refresh.
+        self.getInfo()
+        return self._should_seal
+
+    def seal( self ):
+        '''Mark the sensor for sealing (persistent).'''
+        return self._manager._apiCall( '%s/seal' % ( self.sid, ), POST )
+
+    def unseal( self ):
+        '''Remove the sensor from sealing (persistent).'''
+        return self._manager._apiCall( '%s/seal' % ( self.sid, ), DELETE )
+
+    def getRetainedEventCount( self, startTime, endTime, isDetailed = False ):
+        '''Get the number of events retained for a given sensor between two second epochs.
+
+        Args:
+            startTime (int): time (unix seconds epoch) of the period start.
+            endTime (int): time (unix seconds epoch) of the period end.
+
+        Returns:
+            Event counts.
+        '''
+        return self._manager._apiCall( 'insight/event_count/%s/%s' % ( self._manager._oid, self.sid ), GET, queryParams = {
+            'start' : startTime,
+            'end' : endTime,
+            'is_detailed' : 'true' if isDetailed else 'false',
+        } )
+
     def __str__( self ):
         return self.sid
 
