@@ -20,15 +20,32 @@ class LcConfigException( Exception ):
     pass
 
 class Sync( object ):
-    def __init__( self, oid, apiKey ):
+    '''Sync object to fetch and apply configs to and from organizations.'''
+
+    def __init__( self, oid = None, env = None, manager = None ):
+        '''Create a Sync object.
+
+        Args:
+            oid (str): organization ID to operate on.
+            env (str): environment name to use.
+            manager (limacharlie.Manager): Manager object to use instead.
+        '''
+
+        sys.stderr.write("Sync is now deprecated in favor of Configs...\n")
+        sys.stderr.flush()
+
         self._confVersion = 2
-        self._oid = oid
-        self._apiKey = apiKey
-        self._man = Manager( self._oid, self._apiKey )
+        if manager is None:
+            self._man = Manager( oid = oid, environment = env )
+        else:
+            self._man = manager
 
     def _coreRuleContent( self, rule ):
         return { k : v for k, v in rule.items() if k in ( 'name', 'detect', 'respond', 'namespace' ) }
 
+    def _coreFPContent( self, rule ):
+        return { k : v for k, v in rule.items() if k in ( 'name', 'data' ) }
+
     def _coreOutputContent( self, output ):
         return { k : v for k, v in output.items() if k != 'name' }
 
@@ -59,14 +76,17 @@ class Sync( object ):
 
         return True
 
-    def fetch( self, toConfigFile, isNoRules = False, isNoOutputs = False, isNoIntegrity = False, isNoLogging = False, isNoExfil = False, isNoResources = False ):
+    def fetch( self, toConfigFile, isNoRules = False, isNoFPs = False, isNoOutputs = False, isNoIntegrity = False, isNoLogging = False, isNoExfil = False, isNoResources = False ):
         '''Retrieves the effective configuration in the cloud to a local config file.
 
         Args:
-            toConfigFile (str): the path to the local config file.
+            toConfigFile (str, dict): the path to the local config file or dict where to store config.
         '''
-        toConfigFile = os.path.abspath( toConfigFile )
-        asConf = { 'version' : self._confVersion }
+        if not isinstance( toConfigFile, dict ):
+            toConfigFile = os.path.abspath( toConfigFile )
+            asConf = { 'version' : self._confVersion }
+        else:
+            asConf = toConfigFile
         if not isNoRules:
             rules = {}
             # Check which namespaces we have access to.
@@ -75,6 +95,8 @@ class Sync( object ):
                 availableNamespaces.append( 'general' )
             if self._man.testAuth( permissions = [ 'dr.list.managed' ] ):
                 availableNamespaces.append( 'managed' )
+            if self._man.testAuth( permissions = [ 'dr.list.replicant' ] ):
+                availableNamespaces.append( 'replicant' )
 
             # Fetch the rules from all the namespaces we have access to.
             for namespace in availableNamespaces:
@@ -87,6 +109,13 @@ class Sync( object ):
                     continue
                 rules[ ruleName ] = self._coreRuleContent( rule )
             asConf[ 'rules' ] = rules
+        if not isNoFPs:
+            rules = {}
+            fps = self._man.fps()
+
+            for ruleName, rule in list( fps.items() ):
+                rules[ ruleName ] = self._coreFPContent( rule )
+            asConf[ 'fps' ] = rules
         if not isNoOutputs:
             outputs = self._man.outputs()
             for outputName, output in outputs.items():
@@ -105,38 +134,52 @@ class Sync( object ):
         if not isNoExfil:
             exfilRules = Exfil( self._man ).getRules()
             for ruleName, rule in exfilRules[ 'watch' ].items():
+                if '' == rule[ 'operator' ]:
+                    # This is a [secret] rule, let's not mirror it since
+                    # it is handled by a Service.
+                    continue
                 exfilRules[ 'watch' ][ ruleName ] = self._coreExfilContent( rule )
             for ruleName, rule in exfilRules[ 'list' ].items():
                 exfilRules[ 'list' ][ ruleName ] = self._coreExfilContent( rule )
             asConf[ 'exfil' ] = exfilRules
         if not isNoResources:
             asConf[ 'resources' ] = self._man.getSubscriptions()
-        with open( toConfigFile, 'wb' ) as f:
-            f.write( yaml.safe_dump( asConf, default_flow_style = False ).encode() )
+        if not isinstance( toConfigFile, dict ):
+            with open( toConfigFile, 'wb' ) as f:
+                f.write( yaml.safe_dump( asConf, default_flow_style = False ).encode() )
 
-    def push( self, fromConfigFile, isForce = False, isDryRun = False, isNoRules = False, isNoOutputs = False, isNoIntegrity = False, isNoLogging = False, isNoExfil = False, isNoResources = False ):
+    def push( self, fromConfigFile, isForce = False, isDryRun = False, isNoRules = False, isNoFPs = False, isNoOutputs = False, isNoIntegrity = False, isNoLogging = False, isNoExfil = False, isNoResources = False ):
         '''Apply the configuratiion in a local config file to the effective configuration in the cloud.
 
+        Users should favor using the "push<Type>()" convenience functions instead of the
+        main "push()" function as they are safer to use in the event support for new
+        data-types is added to the "push()" function.
+
         Args:
-            fromConfigFile (str): the path to the config file.
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
             isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
             isDryRun (boolean): if True will only simulate the effect of a push.
 
         Returns:
             a generator of changes as tuple (changeType, dataType, dataName).
         '''
-        fromConfigFile = os.path.abspath( fromConfigFile )
+        if isinstance( fromConfigFile, dict ):
+            # The config is already in memory.
+            asConf = fromConfigFile
+        else:
+            # Load the config file from disk.
+            fromConfigFile = os.path.abspath( fromConfigFile )
 
-        # Config files are always evaluated relative to the current one.
-        contextPath = os.path.dirname( fromConfigFile )
-        currentPath = os.getcwd()
-        os.chdir( contextPath )
+            # Config files are always evaluated relative to the current one.
+            contextPath = os.path.dirname( fromConfigFile )
+            currentPath = os.getcwd()
+            os.chdir( contextPath )
 
-        # This function also does the bulk of the validation.
-        asConf = self._loadEffectiveConfig( fromConfigFile )
+            # This function also does the bulk of the validation.
+            asConf = self._loadEffectiveConfig( fromConfigFile )
 
-        # Revert the previous CWD.
-        os.chdir( currentPath )
+            # Revert the previous CWD.
+            os.chdir( currentPath )
 
         if not isNoRules:
             # Check all the namespaces we have access to.
@@ -145,6 +188,8 @@ class Sync( object ):
                 availableNamespaces.append( 'general' )
             if self._man.testAuth( permissions = [ 'dr.list.managed' ] ):
                 availableNamespaces.append( 'managed' )
+            if self._man.testAuth( permissions = [ 'dr.list.replicant' ] ):
+                availableNamespaces.append( 'replicant' )
 
             # Get the current rules, we will try not to push for no reason.
             currentRules = {}
@@ -165,9 +210,6 @@ class Sync( object ):
                         # Exact same, no point in pushing.
                         yield ( '=', 'rule', ruleName )
                         continue
-                if 'test' == ruleName:
-                    print( currentRules[ ruleName ] )
-                    print( rule )
                 if not isDryRun:
                     if previousNamespace is not None and ruleNamespace != previousNamespace:
                         # Looks like the rule changed namespace.
@@ -187,11 +229,41 @@ class Sync( object ):
                     # Ignore special replicant rules.
                     if ruleName.startswith( '__' ):
                         continue
-                    if ruleName not in asConf[ 'rules' ]:
+                    if ruleName not in asConf.get( 'rules', {} ):
                         if not isDryRun:
                             self._man.del_rule( ruleName, namespace = rule.get( 'namespace', 'general' ) )
                         yield ( '-', 'rule', ruleName )
 
+        if not isNoFPs:
+            # Get the current rules, we will try not to push for no reason.
+            currentRules = { k : self._coreFPContent( v ) for k, v in self._man.fps().items() }
+
+            # Start by adding the rules with isReplace.
+            for ruleName, rule in asConf.get( 'fps', {} ).items():
+                rule = self._coreFPContent( rule )
+                # Check to see if it is already in the current rules and in the right format.
+                if ruleName in currentRules:
+                    if self._isJsonEqual( rule[ 'data' ], currentRules[ ruleName ][ 'data' ] ):
+                        # Exact same, no point in pushing.
+                        yield ( '=', 'fp', ruleName )
+                        continue
+                if not isDryRun:
+                    self._man.add_fp( ruleName, rule[ 'data' ], isReplace = True )
+                yield ( '+', 'fp', ruleName )
+
+            # If we are not told to isForce, this is it.
+            if isForce:
+                currentRules = self._man.fps()
+
+                # Now if isForce was specified, list existing rules and remove the ones
+                # not in our list.
+                for ruleName, rule in currentRules.items():
+                    # Ignore special replicant rules.
+                    if ruleName not in asConf.get( 'fps', {} ):
+                        if not isDryRun:
+                            self._man.del_fp( ruleName )
+                        yield ( '-', 'fp', ruleName )
+
         if not isNoOutputs:
             # Get the current outputs, we will try not to push for no reason.
             currentOutputs = { k : self._coreOutputContent( v ) for k, v in self._man.outputs().items() }
@@ -210,7 +282,7 @@ class Sync( object ):
                 # Now if isForce was specified, list the existing outputs and remove the ones
                 # not in our list.
                 for outputName, output in self._man.outputs().items():
-                    if outputName not in asConf[ 'outputs' ]:
+                    if outputName not in asConf.get( 'outputs', {} ):
                         if not isDryRun:
                             self._man.del_output( outputName )
                         yield ( '-', 'output', outputName )
@@ -236,7 +308,7 @@ class Sync( object ):
                 # Now if isForce was specified, list the existing rules and remove the ones
                 # not in our list.
                 for ruleName, rule in integrityReplicant.getRules().items():
-                    if ruleName not in asConf[ 'integrity' ]:
+                    if ruleName not in asConf.get( 'integrity', {} ):
                         if not isDryRun:
                             integrityReplicant.removeRule( ruleName )
                         yield ( '-', 'integrity', ruleName )
@@ -254,14 +326,17 @@ class Sync( object ):
                     loggingReplicant.addRule( ruleName,
                                               patterns = rule[ 'patterns' ],
                                               tags = rule.get( 'tags', [] ),
-                                              platforms = rule.get( 'platforms', [] ) )
+                                              platforms = rule.get( 'platforms', [] ),
+                                              isDeleteAfter = rule.get( 'is_delete_after', False ),
+                                              isIgnoreCert = rule.get( 'is_ignore_cert', False ), 
+                                              daysRetention = rule.get( 'days_retention', 0 ) )
                 yield ( '+', 'logging', ruleName )
 
             if isForce:
                 # Now if isForce was specified, list the existing rules and remove the ones
                 # not in our list.
                 for ruleName, rule in loggingReplicant.getRules().items():
-                    if ruleName not in asConf[ 'logging' ]:
+                    if ruleName not in asConf.get( 'logging', {} ):
                         if not isDryRun:
                             loggingReplicant.removeRule( ruleName )
                         yield ( '-', 'logging', ruleName )
@@ -302,7 +377,7 @@ class Sync( object ):
                 # Now if isForce was specified, list the existing rules and remove the ones
                 # not in our list.
                 for ruleName, rule in exfilReplicant.getRules().get( 'watch', {} ).items():
-                    if ruleName not in asConf[ 'exfil' ].get( 'watch', {} ):
+                    if ruleName not in asConf.get( 'exfil', {} ).get( 'watch', {} ):
                         if not isDryRun:
                             exfilReplicant.removeWatchRule( ruleName )
                         yield ( '-', 'exfil-watch', ruleName )
@@ -314,7 +389,7 @@ class Sync( object ):
         if not isNoResources:
             currentResources = self._man.getSubscriptions()
             for cat in asConf.get( 'resources', {} ):
-                for resName in asConf[ 'resources' ][ cat ]:
+                for resName in asConf.get( 'resources', {} )[ cat ]:
                     fullResName = '%s/%s' % ( cat, resName )
                     if resName not in currentResources.get( cat, [] ):
                         if not isDryRun:
@@ -365,10 +440,101 @@ class Sync( object ):
 
         return asConf
 
-if __name__ == '__main__':
+    def pushRules( self, fromConfigFile, isForce = False, isDryRun = False ):
+        '''Convenience function to push the D&R rules in a local config file to the effective configuration in the cloud.
+
+        Args:
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
+            isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
+            isDryRun (boolean): if True will only simulate the effect of a push.
+
+        Returns:
+            a generator of changes as tuple (changeType, dataType, dataName).
+        '''
+        return list( self.push( fromConfigFile, isForce = isForce, isDryRun = isDryRun, isNoRules = False, isNoFPs = True, isNoOutputs = True, isNoIntegrity = True, isNoLogging = True, isNoExfil = True, isNoResources = True ) )
+
+    def pushFPs( self, fromConfigFile, isForce = False, isDryRun = False ):
+        '''Convenience function to push the FP rules in a local config file to the effective configuration in the cloud.
+
+        Args:
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
+            isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
+            isDryRun (boolean): if True will only simulate the effect of a push.
+
+        Returns:
+            a generator of changes as tuple (changeType, dataType, dataName).
+        '''
+        return list( self.push( fromConfigFile, isForce = isForce, isDryRun = isDryRun, isNoRules = True, isNoFPs = False, isNoOutputs = True, isNoIntegrity = True, isNoLogging = True, isNoExfil = True, isNoResources = True ) )
+
+    def pushOutputs( self, fromConfigFile, isForce = False, isDryRun = False ):
+        '''Convenience function to push the outputs in a local config file to the effective configuration in the cloud.
+
+        Args:
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
+            isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
+            isDryRun (boolean): if True will only simulate the effect of a push.
+
+        Returns:
+            a generator of changes as tuple (changeType, dataType, dataName).
+        '''
+        return list( self.push( fromConfigFile, isForce = isForce, isDryRun = isDryRun, isNoRules = True, isNoFPs = True, isNoOutputs = False, isNoIntegrity = True, isNoLogging = True, isNoExfil = True, isNoResources = True ) )
+
+    def pushIntegrity( self, fromConfigFile, isForce = False, isDryRun = False ):
+        '''Convenience function to push the Integrity configs in a local config file to the effective configuration in the cloud.
+
+        Args:
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
+            isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
+            isDryRun (boolean): if True will only simulate the effect of a push.
+
+        Returns:
+            a generator of changes as tuple (changeType, dataType, dataName).
+        '''
+        return list( self.push( fromConfigFile, isForce = isForce, isDryRun = isDryRun, isNoRules = True, isNoFPs = True, isNoOutputs = True, isNoIntegrity = False, isNoLogging = True, isNoExfil = True, isNoResources = True ) )
+
+    def pushLogging( self, fromConfigFile, isForce = False, isDryRun = False ):
+        '''Convenience function to push the Logging configs in a local config file to the effective configuration in the cloud.
+
+        Args:
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
+            isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
+            isDryRun (boolean): if True will only simulate the effect of a push.
+
+        Returns:
+            a generator of changes as tuple (changeType, dataType, dataName).
+        '''
+        return list( self.push( fromConfigFile, isForce = isForce, isDryRun = isDryRun, isNoRules = True, isNoFPs = True, isNoOutputs = True, isNoIntegrity = True, isNoLogging = False, isNoExfil = True, isNoResources = True ) )
+
+    def pushExfil( self, fromConfigFile, isForce = False, isDryRun = False ):
+        '''Convenience function to push the Exfil configs in a local config file to the effective configuration in the cloud.
+
+        Args:
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
+            isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
+            isDryRun (boolean): if True will only simulate the effect of a push.
+
+        Returns:
+            a generator of changes as tuple (changeType, dataType, dataName).
+        '''
+        return list( self.push( fromConfigFile, isForce = isForce, isDryRun = isDryRun, isNoRules = True, isNoFPs = True, isNoOutputs = True, isNoIntegrity = True, isNoLogging = True, isNoExfil = False, isNoResources = True ) )
+
+    def pushResources( self, fromConfigFile, isForce = False, isDryRun = False ):
+        '''Convenience function to push the Resources configs in a local config file to the effective configuration in the cloud.
+
+        Args:
+            fromConfigFile (str/dict): the path to the config file or dict of a config file content.
+            isForce (boolean): if True will remove configurations in the cloud that are not present in the local file.
+            isDryRun (boolean): if True will only simulate the effect of a push.
+
+        Returns:
+            a generator of changes as tuple (changeType, dataType, dataName).
+        '''
+        return list( self.push( fromConfigFile, isForce = isForce, isDryRun = isDryRun, isNoRules = True, isNoFPs = True, isNoOutputs = True, isNoIntegrity = True, isNoLogging = True, isNoExfil = True, isNoResources = False ) )
+
+def main( sourceArgs = None ):
     import argparse
 
-    parser = argparse.ArgumentParser( prog = 'limacharlie.io sync' )
+    parser = argparse.ArgumentParser( prog = 'limacharlie sync' )
     parser.add_argument( 'action',
                          type = lambda x: str( x ).lower().strip(),
                          help = 'the action to perform, one of "fetch" or "push".' )
@@ -376,7 +542,14 @@ if __name__ == '__main__':
                          type = lambda x: str( uuid.UUID( x.strip() ) ),
                          required = False,
                          dest = 'oid',
+                         default = None,
                          help = 'the OID to authenticate as, if not specified global creds will be used.' )
+    parser.add_argument( '-e', '--environment',
+                         type = str,
+                         required = False,
+                         dest = 'environment',
+                         default = None,
+                         help = 'the name of the LimaCharlie environment (as defined in ~/.limacharlie) to use, otherwise global creds will be used.' )
     parser.add_argument( '-f', '--force',
                          required = False,
                          default = False,
@@ -395,6 +568,12 @@ if __name__ == '__main__':
                          action = 'store_true',
                          dest = 'isNoRules',
                          help = 'if specified, ignore D&R rules from operations' )
+    parser.add_argument( '--no-fp',
+                         required = False,
+                         default = False,
+                         action = 'store_true',
+                         dest = 'isNoFPs',
+                         help = 'if specified, ignore False Psitive rules from operations' )
     parser.add_argument( '--no-outputs',
                          required = False,
                          default = False,
@@ -431,13 +610,7 @@ if __name__ == '__main__':
                          required = False,
                          dest = 'config',
                          help = 'path to the LCConf file to use' )
-    parser.add_argument( '-k', '--api-key',
-                         type = str,
-                         default = None,
-                         required = False,
-                         dest = 'apiKey',
-                         help = 'path to the file holding your API Key, or "-" to consume it from STDIN' )
-    args = parser.parse_args()
+    args = parser.parse_args( sourceArgs )
 
     if args.isDryRun:
         print( '!!! DRY RUN !!!' )
@@ -452,30 +625,17 @@ if __name__ == '__main__':
     if args.isNoExfil:
         print( '!!! NO EXFIL REPLICANT !!!' )
 
-    if args.apiKey is not None:
-        secretKey = args.apiKey.strip()
-        if '-' == secretKey:
-            print( "Using API Key from STDIN" )
-            if _IS_PYTHON_2:
-                secretKey = raw_input().strip()
-            else:
-                secretKey = input().strip()
-        else:
-            secretKey = os.path.abspath( secretKey )
-            print( "Using API Key in: %s" % secretKey )
-            with open( secretKey, 'rb' ) as f:
-                secretKey = f.read().decode().strip()
-    else:
-        secretKey = None
-
     if args.action not in ( 'fetch', 'push' ):
         print( "Action %s is not supported." % args.action )
         sys.exit( 1 )
 
-    s = Sync( args.oid, secretKey )
+    s = Sync( oid = args.oid, env = args.environment )
 
     if 'fetch' == args.action:
-        s.fetch( args.config, isNoRules = args.isNoRules, isNoOutputs = args.isNoOutputs, isNoIntegrity = args.isNoIntegrity, isNoLogging = args.isNoLogging, isNoExfil = args.isNoExfil, isNoResources = args.isNoResources )
+        s.fetch( args.config, isNoRules = args.isNoRules, isNoFPs = args.isNoFPs, isNoOutputs = args.isNoOutputs, isNoIntegrity = args.isNoIntegrity, isNoLogging = args.isNoLogging, isNoExfil = args.isNoExfil, isNoResources = args.isNoResources )
     elif 'push' == args.action:
-        for modification, category, element in s.push( args.config, isForce = args.isForce, isDryRun = args.isDryRun, isNoRules = args.isNoRules, isNoOutputs = args.isNoOutputs, isNoIntegrity = args.isNoIntegrity, isNoLogging = args.isNoLogging, isNoExfil = args.isNoExfil, isNoResources = args.isNoResources ):
-            print( '%s %s %s' % ( modification, category, element ) )
\ No newline at end of file
+        for modification, category, element in s.push( args.config, isForce = args.isForce, isDryRun = args.isDryRun, isNoRules = args.isNoRules, isNoFPs = args.isNoFPs, isNoOutputs = args.isNoOutputs, isNoIntegrity = args.isNoIntegrity, isNoLogging = args.isNoLogging, isNoExfil = args.isNoExfil, isNoResources = args.isNoResources ):
+            print( '%s %s %s' % ( modification, category, element ) )
+
+if __name__ == '__main__':
+    main()
\ No newline at end of file
